/*
 *  Copyright 2001-present Stephen Colebourne
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.joda.beans.sample;

import java.util.Map;
import java.util.NoSuchElementException;

import org.joda.beans.Bean;
import org.joda.beans.JodaBeanUtils;
import org.joda.beans.MetaBean;
import org.joda.beans.MetaProperty;
import org.joda.beans.Property;
import org.joda.beans.gen.BeanDefinition;
import org.joda.beans.gen.PropertyDefinition;
import org.joda.beans.impl.direct.DirectFieldsBeanBuilder;
import org.joda.beans.impl.direct.DirectMetaBean;
import org.joda.beans.impl.direct.DirectMetaProperty;
import org.joda.beans.impl.direct.DirectMetaPropertyMap;

/**
 * Mock tuple, used for testing.
 * 
 * @author Stephen Colebourne
 */
@BeanDefinition(builderScope = "package", metaScope = "package")
public final class TupleFinal<X, Y, Z> implements Bean {

    /**
     * The first value.
     */
    @PropertyDefinition
    private X first;
    /**
     * The second value.
     */
    @PropertyDefinition
    private Y second;
    /**
     * The third value.
     */
    @PropertyDefinition
    private Z third;

    //------------------------- AUTOGENERATED START -------------------------
    /**
     * The meta-bean for {@code TupleFinal}.
     * @return the meta-bean, not null
     */
    @SuppressWarnings("rawtypes")
    public static TupleFinal.Meta meta() {
        return TupleFinal.Meta.INSTANCE;
    }

    /**
     * The meta-bean for {@code TupleFinal}.
     * @param <R>  the first generic type
     * @param <S>  the second generic type
     * @param <T>  the second generic type
     * @param cls1  the first generic type
     * @param cls2  the second generic type
     * @param cls3  the third generic type
     * @return the meta-bean, not null
     */
    @SuppressWarnings("unchecked")
    public static <R, S, T> TupleFinal.Meta<R, S, T> metaTupleFinal(Class<R> cls1, Class<S> cls2, Class<T> cls3) {
        return TupleFinal.Meta.INSTANCE;
    }

    static {
        MetaBean.register(TupleFinal.Meta.INSTANCE);
    }

    /**
     * Returns a builder used to create an instance of the bean.
     * @param <X>  the type
     * @param <Y>  the type
     * @param <Z>  the type
     * @return the builder, not null
     */
    static <X, Y, Z> TupleFinal.Builder<X, Y, Z> builder() {
        return new TupleFinal.Builder<>();
    }

    /**
     * Restricted constructor.
     * @param builder  the builder to copy from, not null
     */
    private TupleFinal(TupleFinal.Builder<X, Y, Z> builder) {
        this.first = builder.first;
        this.second = builder.second;
        this.third = builder.third;
    }

    @SuppressWarnings("unchecked")
    @Override
    public TupleFinal.Meta<X, Y, Z> metaBean() {
        return TupleFinal.Meta.INSTANCE;
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the first value.
     * @return the value of the property
     */
    public X getFirst() {
        return first;
    }

    /**
     * Sets the first value.
     * @param first  the new value of the property
     */
    public void setFirst(X first) {
        this.first = first;
    }

    /**
     * Gets the the {@code first} property.
     * @return the property, not null
     */
    public Property<X> first() {
        return metaBean().first().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the second value.
     * @return the value of the property
     */
    public Y getSecond() {
        return second;
    }

    /**
     * Sets the second value.
     * @param second  the new value of the property
     */
    public void setSecond(Y second) {
        this.second = second;
    }

    /**
     * Gets the the {@code second} property.
     * @return the property, not null
     */
    public Property<Y> second() {
        return metaBean().second().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the third value.
     * @return the value of the property
     */
    public Z getThird() {
        return third;
    }

    /**
     * Sets the third value.
     * @param third  the new value of the property
     */
    public void setThird(Z third) {
        this.third = third;
    }

    /**
     * Gets the the {@code third} property.
     * @return the property, not null
     */
    public Property<Z> third() {
        return metaBean().third().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Returns a builder that allows this bean to be mutated.
     * @return the mutable builder, not null
     */
    Builder<X, Y, Z> toBuilder() {
        return new Builder<>(this);
    }

    @Override
    public TupleFinal<X, Y, Z> clone() {
        return JodaBeanUtils.cloneAlways(this);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj != null && obj.getClass() == this.getClass()) {
            TupleFinal<?, ?, ?> other = (TupleFinal<?, ?, ?>) obj;
            return JodaBeanUtils.equal(getFirst(), other.getFirst()) &&
                    JodaBeanUtils.equal(getSecond(), other.getSecond()) &&
                    JodaBeanUtils.equal(getThird(), other.getThird());
        }
        return false;
    }

    @Override
    public int hashCode() {
        int hash = getClass().hashCode();
        hash = hash * 31 + JodaBeanUtils.hashCode(getFirst());
        hash = hash * 31 + JodaBeanUtils.hashCode(getSecond());
        hash = hash * 31 + JodaBeanUtils.hashCode(getThird());
        return hash;
    }

    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder(128);
        buf.append("TupleFinal{");
        buf.append("first").append('=').append(JodaBeanUtils.toString(getFirst())).append(',').append(' ');
        buf.append("second").append('=').append(JodaBeanUtils.toString(getSecond())).append(',').append(' ');
        buf.append("third").append('=').append(JodaBeanUtils.toString(getThird()));
        buf.append('}');
        return buf.toString();
    }

    //-----------------------------------------------------------------------
    /**
     * The meta-bean for {@code TupleFinal}.
     * @param <X>  the type
     * @param <Y>  the type
     * @param <Z>  the type
     */
    static final class Meta<X, Y, Z> extends DirectMetaBean {
        /**
         * The singleton instance of the meta-bean.
         */
        @SuppressWarnings("rawtypes")
        static final Meta INSTANCE = new Meta();

        /**
         * The meta-property for the {@code first} property.
         */
        @SuppressWarnings({"unchecked", "rawtypes" })
        private final MetaProperty<X> first = (DirectMetaProperty) DirectMetaProperty.ofReadWrite(
                this, "first", TupleFinal.class, Object.class);
        /**
         * The meta-property for the {@code second} property.
         */
        @SuppressWarnings({"unchecked", "rawtypes" })
        private final MetaProperty<Y> second = (DirectMetaProperty) DirectMetaProperty.ofReadWrite(
                this, "second", TupleFinal.class, Object.class);
        /**
         * The meta-property for the {@code third} property.
         */
        @SuppressWarnings({"unchecked", "rawtypes" })
        private final MetaProperty<Z> third = (DirectMetaProperty) DirectMetaProperty.ofReadWrite(
                this, "third", TupleFinal.class, Object.class);
        /**
         * The meta-properties.
         */
        private final Map<String, MetaProperty<?>> metaPropertyMap$ = new DirectMetaPropertyMap(
                this, null,
                "first",
                "second",
                "third");

        /**
         * Restricted constructor.
         */
        private Meta() {
        }

        @Override
        protected MetaProperty<?> metaPropertyGet(String propertyName) {
            switch (propertyName.hashCode()) {
                case 97440432:  // first
                    return first;
                case -906279820:  // second
                    return second;
                case 110331239:  // third
                    return third;
            }
            return super.metaPropertyGet(propertyName);
        }

        @Override
        public TupleFinal.Builder<X, Y, Z> builder() {
            return new TupleFinal.Builder<>();
        }

        @SuppressWarnings({"unchecked", "rawtypes" })
        @Override
        public Class<? extends TupleFinal<X, Y, Z>> beanType() {
            return (Class) TupleFinal.class;
        }

        @Override
        public Map<String, MetaProperty<?>> metaPropertyMap() {
            return metaPropertyMap$;
        }

        //-----------------------------------------------------------------------
        /**
         * The meta-property for the {@code first} property.
         * @return the meta-property, not null
         */
        public MetaProperty<X> first() {
            return first;
        }

        /**
         * The meta-property for the {@code second} property.
         * @return the meta-property, not null
         */
        public MetaProperty<Y> second() {
            return second;
        }

        /**
         * The meta-property for the {@code third} property.
         * @return the meta-property, not null
         */
        public MetaProperty<Z> third() {
            return third;
        }

        //-----------------------------------------------------------------------
        @Override
        protected Object propertyGet(Bean bean, String propertyName, boolean quiet) {
            switch (propertyName.hashCode()) {
                case 97440432:  // first
                    return ((TupleFinal<?, ?, ?>) bean).getFirst();
                case -906279820:  // second
                    return ((TupleFinal<?, ?, ?>) bean).getSecond();
                case 110331239:  // third
                    return ((TupleFinal<?, ?, ?>) bean).getThird();
            }
            return super.propertyGet(bean, propertyName, quiet);
        }

        @SuppressWarnings("unchecked")
        @Override
        protected void propertySet(Bean bean, String propertyName, Object newValue, boolean quiet) {
            switch (propertyName.hashCode()) {
                case 97440432:  // first
                    ((TupleFinal<X, Y, Z>) bean).setFirst((X) newValue);
                    return;
                case -906279820:  // second
                    ((TupleFinal<X, Y, Z>) bean).setSecond((Y) newValue);
                    return;
                case 110331239:  // third
                    ((TupleFinal<X, Y, Z>) bean).setThird((Z) newValue);
                    return;
            }
            super.propertySet(bean, propertyName, newValue, quiet);
        }

    }

    //-----------------------------------------------------------------------
    /**
     * The bean-builder for {@code TupleFinal}.
     * @param <X>  the type
     * @param <Y>  the type
     * @param <Z>  the type
     */
    static final class Builder<X, Y, Z> extends DirectFieldsBeanBuilder<TupleFinal<X, Y, Z>> {

        private X first;
        private Y second;
        private Z third;

        /**
         * Restricted constructor.
         */
        private Builder() {
        }

        /**
         * Restricted copy constructor.
         * @param beanToCopy  the bean to copy from, not null
         */
        private Builder(TupleFinal<X, Y, Z> beanToCopy) {
            this.first = beanToCopy.getFirst();
            this.second = beanToCopy.getSecond();
            this.third = beanToCopy.getThird();
        }

        //-----------------------------------------------------------------------
        @Override
        public Object get(String propertyName) {
            switch (propertyName.hashCode()) {
                case 97440432:  // first
                    return first;
                case -906279820:  // second
                    return second;
                case 110331239:  // third
                    return third;
                default:
                    throw new NoSuchElementException("Unknown property: " + propertyName);
            }
        }

        @SuppressWarnings("unchecked")
        @Override
        public Builder<X, Y, Z> set(String propertyName, Object newValue) {
            switch (propertyName.hashCode()) {
                case 97440432:  // first
                    this.first = (X) newValue;
                    break;
                case -906279820:  // second
                    this.second = (Y) newValue;
                    break;
                case 110331239:  // third
                    this.third = (Z) newValue;
                    break;
                default:
                    throw new NoSuchElementException("Unknown property: " + propertyName);
            }
            return this;
        }

        @Override
        public Builder<X, Y, Z> set(MetaProperty<?> property, Object value) {
            super.set(property, value);
            return this;
        }

        @Override
        public TupleFinal<X, Y, Z> build() {
            return new TupleFinal<>(this);
        }

        //-----------------------------------------------------------------------
        /**
         * Sets the first value.
         * @param first  the new value
         * @return this, for chaining, not null
         */
        public Builder<X, Y, Z> first(X first) {
            this.first = first;
            return this;
        }

        /**
         * Sets the second value.
         * @param second  the new value
         * @return this, for chaining, not null
         */
        public Builder<X, Y, Z> second(Y second) {
            this.second = second;
            return this;
        }

        /**
         * Sets the third value.
         * @param third  the new value
         * @return this, for chaining, not null
         */
        public Builder<X, Y, Z> third(Z third) {
            this.third = third;
            return this;
        }

        //-----------------------------------------------------------------------
        @Override
        public String toString() {
            StringBuilder buf = new StringBuilder(128);
            buf.append("TupleFinal.Builder{");
            int len = buf.length();
            toString(buf);
            if (buf.length() > len) {
                buf.setLength(buf.length() - 2);
            }
            buf.append('}');
            return buf.toString();
        }

        protected void toString(StringBuilder buf) {
            buf.append("first").append('=').append(JodaBeanUtils.toString(first)).append(',').append(' ');
            buf.append("second").append('=').append(JodaBeanUtils.toString(second)).append(',').append(' ');
            buf.append("third").append('=').append(JodaBeanUtils.toString(third)).append(',').append(' ');
        }

    }

    //-------------------------- AUTOGENERATED END --------------------------
}
